/* libunwind - a platform-independent unwind library
   Copyright (C) 2008 CodeSourcery

This file is part of libunwind.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */

#include "unwind_i.h"
#include "offsets.h"
#include <stdio.h>


int
unw_is_signal_frame (unw_cursor_t *cursor)
{
  struct cursor *c = (struct cursor *) cursor;
  unw_word_t w1, ip;
  unw_addr_space_t as;
  unw_accessors_t *a;
  void *arg;
  int ret;

  as = c->dwarf.as;
  a = unw_get_accessors_int (as);
  arg = c->dwarf.as_arg;

  ip = c->dwarf.ip;

  /* syscall */
  if ((ret = (*a->access_mem) (as, ip + 4, &w1, 0, arg)) < 0)
    return 0;

	if ( w1  == 0x2000083f81f0067UL)
		c->sigcontext_format = SW_SCF_LINUX_SIGFRAME;
	else{
  		if ( w1  == 0x2000083f81f015fUL)
			c->sigcontext_format = SW_SCF_LINUX_RT_SIGFRAME;
		else
			c->sigcontext_format = SW_SCF_NONE;
	}
      	return c->sigcontext_format > SW_SCF_NONE;

  //return -UNW_ENOINFO;
}
#if 0
int
unw_handle_signal_frame (unw_cursor_t *cursor)
{
  int i;
  struct cursor *c = (struct cursor *) cursor;
  unw_word_t sc_addr, sp, sp_addr = c->dwarf.cfa;
  struct dwarf_loc sp_loc = DWARF_LOC (sp_addr, 0);
  int ret;

  if ((ret = dwarf_get (&c->dwarf, sp_loc, &sp)) < 0)
    return -UNW_EUNSPEC;

	
  /* Save the SP and PC to be able to return execution at this point
     later in time (unw_resume).  */
  c->sigcontext_sp = c->dwarf.cfa;
  c->sigcontext_pc = c->dwarf.ip;
	if ( c->sigcontext_format == SW_SCF_LINUX_RT_SIGFRAME) 
	{
  	c->ucontext_addr = sp_addr + sizeof (siginfo_t);
		c->sigcontext_addr = sp_addr + sizeof (siginfo_t) + 48;
	}	
	else
	{
		c->ucontext_addr = 0;
		c->sigcontext_addr = sp_addr;
	}
  sc_addr = c->sigcontext_addr + 16;
  c->dwarf.loc[UNW_SW_64_PC] = DWARF_LOC (sc_addr, 0);

  /* Update the dwarf cursor.
     Set the location of the registers to the corresponding addresses */

	sc_addr = c->sigcontext_addr + 32;

#define  SC_REG_OFFSET(X)   (8 * X)

  for (i = UNW_SW_64_R0; i <= UNW_SW_64_R31; i++)
    {
      c->dwarf.loc[i] = DWARF_LOC (sc_addr + SC_REG_OFFSET(i), 0);
    }
	sc_addr += sizeof(long); //sc_fpregs

  for (i = UNW_SW_64_F0; i < UNW_SW_64_PC; i++)
    {
      c->dwarf.loc[i] = DWARF_LOC (sc_addr + SC_REG_OFFSET(i), 0);
    }

  dwarf_get (&c->dwarf, c->dwarf.loc[UNW_SW_64_R30], &c->dwarf.cfa);
  dwarf_get (&c->dwarf, c->dwarf.loc[UNW_SW_64_PC], &c->dwarf.ip);

  return 1;
}
#endif
